/*
 * Placed into the Public Domain.
 * written by Walter Bright
 * www.digitalmars.com
 */

import object;
import std.c.stdio;
import std.c.string;
import std.c.stdlib;
import std.string;
import std.file;
import std.path;
import std.symtable;

extern (C) void _STI_monitor_staticctor();
extern (C) void _STD_monitor_staticdtor();
extern (C) void _STI_critical_init();
extern (C) void _STD_critical_term();
extern (C) void gc_init();
extern (C) void gc_term();
extern (C) void _minit();
extern (C) void _moduleCtor();
extern (C) void _moduleDtor();
extern (C) void _moduleUnitTests();

extern (C) bool no_catch_exceptions;

version (linux) {
    extern (C) {
	struct sigaction_t {
	    union {
		void *sa_handler;
		void *sa_sigaction;
	    }
	    uint[32] sa_mask;
	    int sa_flags;
	    void *sa_restorer;
	}
	int sigaction(int signum, sigaction_t *act, sigaction_t *oldact);
	extern (C) void __backtrace_handler(int id, void *si, void *uc);
	const int SA_NOCLDSTOP = 1;
	const int SA_NOCLDWAIT = 2;
	const int SA_SIGINFO = 4;
	const int SA_ONSTACK = 0x08000000;
	const int SA_RESTART = 0x10000000;
	const int SA_NODEFER = 0x40000000;
	const int SA_RESETHAND = 0x80000000;
	const int SA_INTERRUPT = 0x20000000;
    }
}

/***********************************
 * The D main() function supplied by the user's program
 */
int main(char[][] args);

/***********************************
 * Substitutes for the C main() function.
 * It's purpose is to wrap the call to the D main()
 * function and catch any unhandled exceptions.
 */

extern (C) int main(int argc, char **argv)
{
    char[] *am;
    char[][] args;
    int i;
    int result;
    int myesp;
    int myebx;

    version (linux)
    {
	_STI_monitor_staticctor();
	_STI_critical_init();
	gc_init();
	am = cast(char[] *) malloc(argc * (char[]).sizeof);
	// BUG: alloca() conflicts with try-catch-finally stack unwinding
	//am = (char[] *) alloca(argc * (char[]).sizeof);

	sigaction_t sa;
	sa.sa_sigaction = &__backtrace_handler;
	sa.sa_flags = SA_SIGINFO;
	sigaction(1, &sa, null);
	sigaction(2, &sa, null);
	sigaction(3, &sa, null);
	sigaction(4, &sa, null);
	sigaction(6, &sa, null);
	sigaction(8, &sa, null);
	sigaction(11, &sa, null);
	sigaction(13, &sa, null);
	sigaction(15, &sa, null);
    }
    version (Win32)
    {
	gc_init();
	_minit();
	am = cast(char[] *) malloc(argc * (char[]).sizeof);
     }

    if (no_catch_exceptions)
    {
	_moduleCtor();
	_moduleUnitTests();

	for (i = 0; i < argc; i++)
	{
	    int len = strlen(argv[i]);
	    am[i] = argv[i][0 .. len];
	}

	args = am[0 .. argc];

	result = main(args);
	_moduleDtor();
	gc_term();
    }
    else
    {
	try
	{
	    _moduleCtor();
	    _moduleUnitTests();

	    for (i = 0; i < argc; i++)
	    {
		int len = strlen(argv[i]);
		am[i] = argv[i][0 .. len];
	    }

	    args = am[0 .. argc];

	    result = main(args);
	    _moduleDtor();
	    gc_term();
	}
	catch (Object o)
	{
	    printf("Error: ");
	    Exception e = cast(Exception)o;
	    if (e) {
		if (Symtable.getDefaultSymtable()) {
		    e.print(Symtable.getDefaultSymtable());
		    exit(EXIT_FAILURE);
		}
		version (linux) {
		    char[] nmfile = addExt(args[0], "nm");
		    if (exists(nmfile)) {
			Symtable st =
			    new Symtable(nmfile, SymtableType.BINUTILS_NM);
			e.print(st);
			exit(EXIT_FAILURE);
		    }
		}
		version (Win32) {
		    char[] mapfile = addExt(args[0], "map");
		    if (exists(mapfile)) {
			Symtable st =
			    new Symtable(mapfile, SymtableType.DMC_MAP);
			e.print(st);
			exit(EXIT_FAILURE);
		    }
		}
		Symtable st = new Symtable("", SymtableType.NONE);
		e.print(st);
		exit(EXIT_FAILURE);
	    }
	    else {
		o.print();
	    }
	    exit(EXIT_FAILURE);
	}
    }

    version (linux)
    {
	free(am);
	_STD_critical_term();
	_STD_monitor_staticdtor();
    }
    return result;
}

